/*
% cc -O -o $HOME/bin.`uname -m`/# @
#% gcc -O -o $HOME/bin.`arch`/# @
*/
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#define BUFFERSIZE 2048

int
do_readlink(char *path, char *buf, int len)
{
	int rv;

	while ((rv = readlink(path,buf,len)) < 0 && errno == EAGAIN) ;

	if (rv >= 0 && rv < BUFFERSIZE)
		buf[rv] = 0;

	return rv;
}

int
exists(char *path)
{
	struct stat buf;
	int rv, e = 0;

	while ((rv = stat(path,&buf)) < 0 && (e = errno) == EAGAIN) ;

	return rv >= 0
		|| (e != ENOENT && e != EACCES && e != ELOOP && e != EMULTIHOP
		 && e != ENAMETOOLONG && e != ENOTDIR && e != ENOLINK);
}

int
main(int argc,char *argv[])
{
	char buf[BUFFERSIZE];
	int	cc;
	int	fi = 0;
	int	err = 0;
	int	quiet = 0;
	int errors = 0;
	int	list = 0;
	int	maxlen = 0;
	int	symlinklist = 0;

	if (argv[1] && (argv[1])[0] == '-') {
		if (strchr(argv[1],'q'))
			quiet = 1;
		if (strchr(argv[1],'l'))
			list = 1;
		if (strchr(argv[1],'e'))
			errors = 1;
		if (strchr(argv[1],'s')) {
			symlinklist = 1;
			list = 0;
		}
		if (strchr(argv[1],'?')
		 || strchr(argv[1],'h')) {
			fprintf(stderr,"usage: %s  [-ehlqs?] <files>\n",argv[0]);
			if (strchr(argv[1],'h')) {
				fprintf(stderr,"\
-e: errors; output broken links only\n\
-h: help; output this help\n\
-l: list; list both link and target\n\
-q: quiet; do not output errors for broken links\n\
-s: symlink; output links as link commands (e.g. ln -s foo bar)\n");
				exit(0);
			}
		}
		++fi;
	}
	if (list) {
		int i = fi;
		while (argv[++i]) {
			int len = strlen(argv[i]);
			if (len > maxlen
			 && (!errors || errors && !exists(argv[i]))
			 && do_readlink(argv[i], buf, BUFFERSIZE) >= 0)
				maxlen = len;
		}
	}
	while (argv[++fi]) {
		if ((cc = do_readlink(argv[fi], buf, BUFFERSIZE)) <= 0) {
#if !defined(linux) && !defined(__MACH__)
			extern	char	*sys_errlist[];
#endif

			if (!quiet)
				fprintf(stderr,
					"%s: %s %s\n",
					argv[0],
					argv[fi],
					sys_errlist[errno]);
			err++;
		}
		else if (!errors
			  || errors && !exists(argv[fi])) {
			buf[cc] = '\0';
			if (list)
				printf("%-*s -> %s\n",maxlen,argv[fi],buf);
			else if (symlinklist)
				printf("ln -s %s %s\n",buf,argv[fi]);
			else
				printf("%s\n",buf);
		}
	}
	exit(err);
	return err;
}